/*
 * ----------------------------------------------------------------------------
 *
 * ztdb.h
 * Copyright (C) 2003, 2004  Fabian Pietsch
 *
 * ----------------------------------------------------------------------------
 *
 */

/*
 * zzztel database module
 ******************************************************************************
 * database abstraction interface for zzztel
 *
 */

/* handle multiple inclusions */
#ifndef __ZTDB_H
#define __ZTDB_H

/* need ZTTN & ZTList */
#include "zttn.h"
#include "ztlist.h"

/*
 * data types used for searching for telephone numbers
 */

/* fields-to-be-searched indicator */
enum zts_fields
{
	ZTS_PRE_INT   = 0x01,   /* international prefix */
	ZTS_PRE_LOCAL = 0x02,   /* local prefix         */
	ZTS_LINE      = 0x04,   /* line number          */
	ZTS_COMMENT   = 0x08,   /* e.g., name           */
	ZTS_ALL       = ZTS_PRE_INT | ZTS_PRE_LOCAL | ZTS_LINE | ZTS_COMMENT
};
typedef enum zts_fields ZTSFields;

/* field indicator test macros */
#define ZTS_HAS_PRE_INT(x)   ((x & ZTS_PRE_INT) == ZTS_PRE_INT)
#define ZTS_HAS_PRE_LOCAL(x) ((x & ZTS_PRE_LOCAL) == ZTS_PRE_LOCAL)
#define ZTS_HAS_LINE(x)      ((x & ZTS_LINE) == ZTS_LINE)
#define ZTS_HAS_COMMENT(x)   ((x & ZTS_COMMENT) == ZTS_COMMENT)
#define ZTS_HAS_ALL(x)       ((x & ZTS_ALL) == ZTS_ALL)

/* telephone number search request */
struct ztsearch
{
	ZTTN      *tn;      /* telephone number template       */
	ZTSFields  fields;  /* fields-to-be-searched indicator */
	int        limit;   /* limit on objects affacted       */
};
typedef struct ztsearch ZTSearch;

/* function prototypes */
extern int ztdb_init(void);
extern int ztdb_finish(void);
extern int ztdb_tnquery(const ZTSearch *search);
extern ZTTN *ztdb_tnget(void);
extern ZTList *ztdb_tngetall(void);
extern int ztdb_tnins(const ZTTN *tn);
extern int ztdb_tndel(const ZTTN *tn);
extern int ztdb_tndelall(const ZTSearch *search);

#endif /* __ZTDB_H */
