/*
 * ----------------------------------------------------------------------------
 *
 * zttn.c -- telephone number object for zzztel
 * Copyright (C) 2004  Fabian Pietsch
 *
 * ----------------------------------------------------------------------------
 *
 */

#include "zttn.h"
#include "zzzutils.h"
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* telephone number template duplication routine */
#define ZTTN_TDUP(dst, src, field) \
do { \
	if (src->field == NULL) \
	{ \
		dst->field = NULL; \
		break; \
	} \
	if ((dst->field = strdup(src->field)) == NULL) \
	{ \
		zperror("strdup()"); \
		goto err_free_ ## field; \
	} \
} while (0)

/* create a new telephone number object */
ZTTN *zttn_new(ZTTN *tn_template)
{
	ZTTN *tn;

	/* allocate memory for struct zttelnum */
	if ((tn = malloc(sizeof(ZTTN))) == NULL)
	{
		zperror("malloc()");
		return NULL;
	}

	/* template given? */
	if (tn_template)
	{
		/* copy template data to new object */
		ZTTN_TDUP(tn, tn_template, pre_int);
		ZTTN_TDUP(tn, tn_template, pre_local);
		ZTTN_TDUP(tn, tn_template, line);
		ZTTN_TDUP(tn, tn_template, comment);
	}
	else
	{
		/* initialize object fields to NULL pointers */
		tn->pre_int = NULL;
		tn->pre_local = NULL;
		tn->line = NULL;
		tn->comment = NULL;
	}

	/* return new telephone number object */
	return tn;

err_free_comment:
	free(tn->comment);
err_free_line:
	free(tn->line);
err_free_pre_local:
	free(tn->pre_local);
err_free_pre_int:
	free(tn->pre_int);

	free(tn);
	return NULL;
}

/* deallocate telephone number object */
void zttn_free(ZTTN *tn)
{
	assert(tn != NULL);

	/* free contained data */
	free(tn->pre_int);
	free(tn->pre_local);
	free(tn->line);
	free(tn->comment);

	/* free the object itself */
	free(tn);
}
