/*
 * ----------------------------------------------------------------------------
 *
 * ztui.c -- zzztel user interface
 * Copyright (C) 2003, 2004  Fabian Pietsch
 *
 * ----------------------------------------------------------------------------
 *
 */

#include "zzztel.h"
#include "ztdb.h"
#include "ztui.h"
#include "zzzutils.h"

#include <curses.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


/* private function prototypes */
static int ztui_tns_insert_getmember(char **member, int maxlen, WINDOW *win, const char *desc);
static int ztui_tns_view_update(ZTUIView *view, int full);


#if 0
/* execute input form */
int ztui_form(int y, int x, int h, int w, struct ztform_element **elv, int elc)
{
	PANEL  *pan_form;
	WINDOW *win_form_all, *win_form;
	struct ztform_element *el;
	int wfh, wfw, i, cont = 1;

	/* compute size of the output window and position it as specified */
	getmaxyx(stdscr, wfh, wfw);
	if (h < 1) h = wfh / 2;
	if (w < 1) w = wfw / 2;
	if (y < 0) y = (wfh - h) / 2;
	if (x < 0) x = (wfw - w) / 2;
	win_form_all = newwin(h, w, y, x);

	/* create new panel for the form */
	if ((pan_form = new_panel(win_form_all)) == NULL)
		return -1;

	/* draw box around panel's border & get subwindow */
	/* wclear(win_form_all); */
	box(win_form_all, 0, 0);
	win_form = derwin(win_form_all, h - 2, w - 4, 1, 2);
	scrollok(win_form, TRUE);
	keypad(win_form, TRUE);

	/* draw elements */
	for (i=0; i<elc; i++)
	{
		el = elv[i];
		wmove(win_form, el->y, el->x);

		switch (el->type) {
		case ZTFET_LABEL:
			waddstr(win_form, el->data);
			break;
		case ZTFET_INPUT:
			whline(win_form, '_', el->w);
			break;
		default:
			delwin(win_form);
			del_panel(pan_form);
			ztui_refresh();
			return -2;
		}
	}

	/* enter event loop */
	while (cont)
	{
		int c;

		/* update screen */
		wcursyncup(win_form);
		wsyncup(win_form);
		ztui_refresh();

		/* read & process user input */
		switch (c = wgetch(win_form))
		{
			case '\r':
			case '\e':
			case KEY_ENTER:
			case KEY_CANCEL:
				cont = 0;
				break;
			default:
				beep();
				break;
		}
	}

	/* free used ressources */
	delwin(win_form);
	del_panel(pan_form);
	delwin(win_form_all);

	/* success */
	return 0;
}

/* show set of telephone numbers */
int ztui_tns_show(const ZTSearch *search)
{
	ZTUITNList  tnlist;
	ZTTN       *tn;
	WINDOW     *win;
	int         wma_y, wma_x, rows, cont = 1;

	/* create window for output and an associated panel */
	tnlist.win_full = newwin(0, 0, 0, 0);
	tnlist.pan_full = new_panel(tnlist.win_full);
	box(tnlist.win_full, 0, 0);

	/* create subwindow for actual output */
	getmaxyx(tnlist.win_full, wma_y, wma_x);
	tnlist.win_inner = derwin(tnlist.win_full, wma_y - 2, wma_x - 4, 1, 2);
	win = tnlist.win_inner;
	/* wclear(tnlist.win_inner); */
	/* wmove(tnlist.win_inner, 0, 0); */
	scrollok(win, TRUE);
	keypad(win, TRUE);

	/* search for telephone numbers */
	if ((rows = ztdb_tnquery(search)) < 0)
		/* an error occurred */
		waddstr(win, "Error in search.\n");
	else if (rows == 0)
		/* result set is empty */
		waddstr(win, "No matches found.\n");
	else
	{
		/* success: output result set */
		wprintw(win, "Found %d telephone numbers:\n\n", rows);
		while ((tn = ztdb_tnget()))
		{
			wprintw(win, "+%s %s %s\t%s\n",
			        tn->pre_int, tn->pre_local, tn->line, tn->comment);
			free(tn);
		}
	}

	/* enter event loop */
	while (cont)
	{
		int c;

		/* update screen */
		wcursyncup(win);
		wsyncup(win);
		ztui_refresh();

		/* read & process user input */
		switch (c = wgetch(win))
		{
			case 'q':
			case '\e':
			case KEY_CANCEL:
				cont = 0;
				break;
			default:
				beep();
				break;
		}
	}
	
	/* free used ressources */
	delwin(tnlist.win_inner);
	del_panel(tnlist.pan_full);
	delwin(tnlist.win_full);

	/* success */
	return 0;
}
#endif

/* refresh terminal */
int ztui_refresh()
{
	update_panels();
	return -(doupdate() == ERR);
}

/* insert new telephone number */
int ztui_tns_insert(int x, int y, int width, int height, ZTTN **tncopy)
{
	WINDOW *full_win;
	PANEL  *full_pan;
	WINDOW *inner_win;
	int     inner_height, inner_width;
	int     result, ret = 0;
	ZTTN    tn;

	/* set up telephone number data structure */
	tn.pre_int = NULL;
	tn.pre_local = NULL;
	tn.line = NULL;
	tn.comment = NULL;

	/* create window for output and an associated panel */
	full_win = newwin(height, width, y, x);
	full_pan = new_panel(full_win);
	box(full_win, 0, 0);

	/* create subwindow for scrolling region */
	inner_height = height - 2;
	inner_width  = width  - 4;
	inner_win = derwin(full_win, inner_height, inner_width, 1, 2);

	/* set up inner window for output */
	/* wclear(inner_win); */
	/* wmove(inner_win, 0, 0); */
	keypad(inner_win, TRUE);
	scrollok(inner_win, TRUE);

	/* update screen */
	wcursyncup(inner_win);
	wsyncup(inner_win);
	ztui_refresh();

	/*
	 * get user input for telephone number to add
	 */

#define ZTUI_TNS_INSERT_GETMEMBER_WRAPPER(member, maxlen, win, desc) \
	do { \
		while ((result = ztui_tns_insert_getmember(member, maxlen, win, desc)) == 1) \
		{ \
			/* FIXME: ask user for insert abortion */ \
			ret = 1; \
			goto out_free; \
		} \
		if (result < 0) \
		{ \
			/*propagate input error */ \
			ret = -1; \
			goto out_free; \
		} \
	} while (0)

	ZTUI_TNS_INSERT_GETMEMBER_WRAPPER(&tn.pre_int, 2, inner_win, "International prefix");
	ZTUI_TNS_INSERT_GETMEMBER_WRAPPER(&tn.pre_local, 5, inner_win, "Local prefix");
	ZTUI_TNS_INSERT_GETMEMBER_WRAPPER(&tn.line, 10, inner_win, "Line's local number");
	ZTUI_TNS_INSERT_GETMEMBER_WRAPPER(&tn.comment, 255, inner_win, "Comment");

#undef ZTUI_TNS_INSERT_GETMEMBER_WRAPPER

	/* insert telephone number into database */
	if ((result = ztdb_tnins(&tn)) < 0)
	{
		/* FIXME: needs to be ncursed */
		zerror("error inserting telephone number into database");
		ret = result;
		goto out_free;
	}

	/* check for additional error possibilities */
	if (result != 1)
	{
		/* FIXME: needs to be ncursed */
		zerror("unexpected result from database insertion (%d != 1)", result);
		ret = -1;
		goto out_free;
	}

	/* leave a copy of the newly inserted tn, if requested */
	if (tncopy)
		if ((*tncopy = zttn_new(&tn)) == NULL)
		{
			ret = -1;
			goto out_free;
		}
	
out_free:
	/* free tn members */
	if (tn.pre_int)
		free(tn.pre_int);
	if (tn.pre_local)
		free(tn.pre_local);
	if (tn.line)
		free(tn.line);
	if (tn.comment)
		free(tn.comment);

	/* free used ncurses ressources */
	delwin(inner_win);
	del_panel(full_pan);
	delwin(full_win);

	/* return success/error indicator */
	return ret;
}

/* get user input for telephone number data structure member */
static int ztui_tns_insert_getmember(char **member, int maxlen, WINDOW *win, const char *desc)
{
	char buf[maxlen + 1];
	int result;

	/* output prompt & read in answer */
	wprintw(win, "%s: ", desc);
	echo();
	result = wgetnstr(win, buf, maxlen);
	noecho();
       
	/* check for input error */
	if (result == ERR)
		return -1;

	/* is answer the empty string? */
	if (buf[0] == '\0')
	{
		/* NULL-ify member & indicate success/empty answer */
		*member = NULL;
		return 1;
	}

	/* copy answer to member */
	if ((*member = strdup(buf)) == NULL)
	{
		zperror("malloc()");
		return -1;
	}

	/* indicate success/non-empty answer */
	return 0;
}

/* search for telephone numbers (search dialog) */
int ztui_tns_search()
{
	ZTSearch  search;
	ZTTN      tn;
	ZTList   *result;
#if 0
	struct ztform_element fel_label = { ZTFET_LABEL, 0, 0, 0, 0, "Test\n" };
	struct ztform_element fel_input = { ZTFET_INPUT, 1, 4, 0, 8, NULL };
	struct ztform_element *fels[] =
	{
		&fel_label,
		&fel_input
	};
#endif
	int rows;

	/* present an input dialog to user */
	/* FIXME */
	/* ztui_form(-1, -1, 0, 0, fels, 2); */

	tn.line = "567890";
	search.tn = &tn;
	/* search.fields = ZTS_LINE; */
	search.fields = 0;
	search.limit = 0;

	/* FIXME: change following addstr() occurences to some sort of message box */
	/* search for telephone numbers */
	/* if ((rows = ztdb_tnquery(&search)) < 0) */
	if ((rows = ztdb_tnquery(NULL)) < 0)
	{
		/* an error occurred */
		addstr("Error in search.\n");
		return 1;
	}

	if (rows == 0)
	{
		/* result set is empty */
		addstr("No matches found.\n");
		return 1;
	}

	/* success: fetch search result */
	if ((result = ztdb_tngetall()) == NULL)
	{
		addstr("Can't fetch search result.\n");
		return 1;
	}

	/* let the user browse the search result */
	return ztui_tns_view(result);
}

/* view list of telephone numbers */
int ztui_tns_view(ZTList *list)
{
	ZTUIView  view;
	int       cont = 1, ret = 0, update_full = 1;

	assert(list != NULL);

	/* initialize view data structure */
	view.list = list;
	view.offset = 0;
	view.pos = 0;
	view.prev_offset = -1;
	view.prev_pos = -1;

	/* create window for output and an associated panel */
	view.full_win = newwin(0, 0, 0, 0);
	view.full_pan = new_panel(view.full_win);
	box(view.full_win, 0, 0);

	/* create subwindow for scrolling region */
	getmaxyx(view.full_win, view.full_maxy, view.full_maxx);
	view.inner_maxy = view.full_maxy - 4;
	view.inner_maxx = view.full_maxx - 4;
	view.inner_win = derwin(view.full_win,
	                        view.inner_maxy, view.inner_maxx, 3, 2);

	/* set up inner window for output */
	/* wclear(view.inner_win); */
	/* wmove(view.inner_win, 0, 0); */
	keypad(view.inner_win, TRUE);
	mvwprintw(view.full_win, 1, 2, "Found %d telephone numbers:", view.list->count);

	/* enter event loop */
	while (cont)
	{
		ZTListEl *tnlistel;
		ZTTN     *tn;
		int       c, result;

		/* update screen */
		if (ztui_tns_view_update(&view, update_full))
		{
			wprintw(view.inner_win, "couldn't update screen\n");
			ztui_refresh();
			ret = -1;
			goto out_free;
		}
		update_full = 0;

		/* read & process user input */
		switch (c = wgetch(view.inner_win))
		{
			case KEY_CANCEL:
			case 'q':
			case '\e':
				/* leave tn list view */
				cont = 0;
				break;

			case KEY_NPAGE:
			case 'D' - 'A' + 1:  /* ^D */
				/* scroll down half a page */
				if (view.offset < ZTUI_VOFFSET_MAX(view))
				{
#if 0
					view.offset =
					  MIN(view.offset + view.inner_maxy / 2,
					      ZTUI_VOFFSET_MAX(view));
#endif
					view.offset += view.inner_maxy / 2;
					view.pos += view.inner_maxy / 2;
				}
				else
					beep();
				break;

			case KEY_PPAGE:
			case 'U' - 'A' + 1:  /* ^U */
				/* scroll up half a page */
				if (view.offset > ZTUI_VOFFSET_MIN(view))
				{
#if 0
					view.offset =
					  MAX(view.offset - view.inner_maxy / 2,
					      ZTUI_VOFFSET_MIN(view));
#endif
					view.offset -= view.inner_maxy / 2;
					view.pos -= view.inner_maxy / 2;
				}
				else
					beep();
				break;

			case KEY_DOWN:
			case 'j':
				/* move selection down */
				if (view.pos < view.list->count - 1)
				{
					view.pos++;

					/* scroll if necessary */
					if (view.pos >= view.offset + view.inner_maxy)
						view.offset++;
				}
				else
					beep();
				break;

			case KEY_UP:
			case 'k':
				/* move selection up */
				if (view.pos > 0)
				{
					view.pos--;

					/* scroll if necessary */
					if (view.pos < view.offset)
						view.offset--;
				}
				else
					beep();
				break;

			case KEY_DC:
			case 'd':
				/* delete telephone number */
				/* FIXME: ask user for confirmation */
				if ((tnlistel = ztlist_idx_to_ptr(view.list, view.pos)) == NULL)
				{
					/* FIXME: has to be ncursed */
					zerror("couldn't transform list index to element pointer");
					break;
				}
				if (ztdb_tndel(tnlistel->tn))
				{
					/* FIXME: has to be ncursed */
					zerror("couldn't delete telephone number");
					break;
				}

				/* delete element from list */
				if (ztlist_del(view.list, tnlistel))
				{
					/* FIXME: has to be ncursed */
					zerror("couldn't delete list element");
					/* telephone number has been deleted in the database,
					 * so abort now to avoid further inconsistency */
					ret = -1;
					goto out_free;
				}

				/* adjust selection */
				if (view.pos == view.list->count - 1)
					view.pos--;
				update_full = 1;

				break;

			case 'i':
				/* insert telephone number into database */
				tn = NULL;
				if ((result = ztui_tns_insert(
				       3, 3, view.full_maxx - 6, view.full_maxy - 6,
				       &tn)) < 0)
				{
					/* error in underlying library calls */
					beep();
					break;
				}

				/* check for successful insertion */
				if (result == 0)
				{
					/* FIXME: somehow a full refresh of the
					 * list might be needed, in case of being
					 * just a search result; store query in list?
					 */

					/* is the tn's copy really there? */
					assert(tn != NULL);
#if 0
					if (!tn)
					{
						/* indicate error (most probably no memory left) */
						ret = -1;
						goto out_free;
					}
#endif

					/* add newly inserted tn to list */
					if (ztlist_add(view.list, tn) == NULL)
					{
						ret = -1;
						goto out_free;
					}
					/* FIXME: maybe add prev_count to ZTView? */

					/* refresh virtual & real screen on next input loop iteration */
					update_full = 1;
				}
				else
					/* refresh real screen now after input window has gone */
					ztui_refresh();

				break;

			default:
				beep();
				break;
		}

		/* fix variables being out-of-range */
		if (view.pos < 0)
			view.pos = 0;
		if (view.pos >= view.list->count)
			view.pos = view.list->count - 1;
		if (view.offset < 0)
			view.offset = 0;
		if (view.offset >= view.list->count)
			view.offset = view.list->count - 1;

		assert((view.offset <= view.pos) &&
		       (view.pos < view.offset + view.inner_maxy));
	}
	
out_free:
	/* free used ressources */
	delwin(view.inner_win);
	del_panel(view.full_pan);
	delwin(view.full_win);

	/* return success/error indicator */
	return ret;
}

/* ztui_tns_view_update() -- update screen in ztui_tns_view() */
static int ztui_tns_view_update(ZTUIView *view, int full)
{
	ZTListEl *tnlistel;
	ZTTN     *tn;
	WINDOW   *inner_win = view->inner_win;
	int       i = 0;

	assert(inner_win != NULL);

	/* do we really need to do anything? */
	if (!full &&
	    (view->offset == view->prev_offset) &&
	    (view->pos    == view->prev_pos))
		return 0;

	/* full update needed? */
	if (view->offset != view->prev_offset)
		full = 1;

	/* clear inner (scrolling / data) window */
	if (full)
		wclear(inner_win);

	/* iterate through list of telephone numbers */
	tnlistel = view->list->first;
	while (tnlistel && (i - view->offset < view->inner_maxy))
	{
		/* is this tn currently user-visible and
		 * is outputting it necessary? */
		if ((i >= view->offset) &&
		    (full ||
		     (i == view->pos) ||
		     (i == view->prev_pos)))
		{
			/* highlight selected tn */
			if (i == view->pos)
				wattron(inner_win, A_STANDOUT);

			/* output this tn */
			tn = tnlistel->tn;
			mvwprintw(inner_win, i - view->offset, 0, "+%s %s %s\t%s",
			          tn->pre_int, tn->pre_local, tn->line, tn->comment);

			/* end highlighting of selected tn */
			if (i == view->pos)
				wattroff(inner_win, A_STANDOUT);
		}

		/* advance to next tn object in linked list */
		tnlistel = tnlistel->next;
		i++;
	}

	/* put cursor on selected tn */
	wmove(inner_win, view->pos - view->offset, 0);

	/* remember current offset / position for next update */
	view->prev_offset = view->offset;
	view->prev_pos = view->pos;

	/* update screen */
	wcursyncup(inner_win);
	wsyncup(inner_win);
	ztui_refresh();

	/* indicate success */
	return 0;
}

/* user interface event loop */
int ztui_main()
{
	int ret;

	/* initialize curses */
	if (initscr() == NULL)
	{
		zerror("can't initialize curses screen handling library");
		return -1;
	}
	cbreak(); noecho();
	nonl(); /* intrflush(stdscr, FALSE); */ keypad(stdscr, TRUE);

#if 0
	/* initialize panels */
	if ((pan_main = new_panel(stdscr)) == NULL)
	{
		zerror("can't initialize panels");
		endwin();
		return -1;
	}
	win_main = panel_window(pan_main);
	box(win_main, 0, 0);
	wmove(win_main, 1, 1);
	ztui_refresh();
#endif

	/* FIXME: do someting */
	ret = ztui_tns_search();

	/* restore terminal settings */
	endwin();
	putchar('\n');

	/* exit from user interface; return success/error indicator */
	return ret;
}
