/*
 * ----------------------------------------------
 *
 * This software belongs to the zzzutils package.
 *
 * Copyright (c) 2003 Fabian "zzznowman" Pietsch.
 * All rights reserved.
 *
 * ----------------------------------------------
 */

/*
 * zzzutils.h -- configuration and prototyping of common functions
 *
 *****************************************************************
 *
 *
 */


/* handle multiple inclusions */
#ifndef __ZZZUTILS_H
#define __ZZZUTILS_H

/*
 * simple macro definitions
 */

/* stringification macros needed for VERSION */
#define Str(x) #x
#define Xstr(x) Str(x)


/*
 * configuration section
 */

/* zzzutils version data */
#define VER_MAJOR 0
#define VER_MINOR 1
#define VER_REVISION 3
#if 0
/* this doesn't work */
#define VERSION "VER_MAJOR.VER_MINOR.VER_REVISION"
#elif 0
/* workaround using double indirection
 * XXX: any hints welcome on how to improve this!
 * [2003-08-22: considering http://www.eskimo.com/~scs/C-faq/q11.17.html
 *  this could be generalized at most, but not really simplified]
 */
#define VERSION__(vermaj, vermin, verrev) #vermaj "." #vermin "." #verrev
#define VERSION_(vermaj, vermin, verrev) VERSION__(vermaj, vermin, verrev)
#define VERSION VERSION_(VER_MAJOR, VER_MINOR, VER_REVISION)
#else /* 0 */
/* generalized implementation 2003-08-22 */
#define VERSION Xstr(VER_MAJOR) "." Xstr(VER_MINOR) "." Xstr(VER_REVISION)
#endif /* 0 */
#define PRODUCT_NAME "zzztel"
#define PRODUCT_VERSION PRODUCT_NAME " " VERSION

/* character used for separating file system path components */
#define PATH_SEPARATOR '/'


/*
 * prototypes for misc.c
 */

void zperror_verbose(const char *file, const int line, const char *fmt, ...);

int zerror_verbose(const char *file, const int line, const char *fmt, ...);
int zwarn_verbose(const char *file, const int line, const char *fmt, ...);

#define zperror(fmt, args...) zperror_verbose(__FILE__, __LINE__, fmt ,##args)
#define zerror(fmt, args...) zerror_verbose(__FILE__, __LINE__, fmt ,##args)
#define zwarn(fmt, args...) zwarn_verbose(__FILE__, __LINE__, fmt ,##args)

#endif /* __ZZZUTILS_H */
